SELECT
  NTILE(2) OVER (ORDER BY 0.5 + (
    CAST(RANDOM() AS REAL) / -1.8446744073709552e+19
  ) ASC NULLS LAST ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) - 1 AS "new_col"
FROM "test" AS "t0"
LIMIT 10