SELECT
  PERCENTILE_CONT(0.25) WITHIN GROUP (ORDER BY
    "t0"."ROW_COUNT") AS "quantile_0_25",
  PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY
    "t0"."ROW_COUNT") AS "quantile_0_75"
FROM "t" AS "t0"